<!--
  ~ Licensed to the Apache Software Foundation (ASF) under one
  ~ or more contributor license agreements.  See the NOTICE file
  ~ distributed with this work for additional information
  ~ regarding copyright ownership.  The ASF licenses this file
  ~ to you under the Apache License, Version 2.0 (the
  ~ "License"); you may not use this file except in compliance
  ~ with the License.  You may obtain a copy of the License at
  ~
  ~     http://www.apache.org/licenses/LICENSE-2.0
  ~
  ~ Unless required by applicable law or agreed to in writing, software
  ~ distributed under the License is distributed on an "AS IS" BASIS,
  ~ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  ~ See the License for the specific language governing permissions and
  ~ limitations under the License.
  -->

<nz-table
  [nzData]="listOfData"
  [nzSize]="'small'"
  [nzFrontPagination]="false"
  [nzShowPagination]="false"
  [nzVirtualItemSize]="36"
  [nzVirtualMaxBufferPx]="300"
  [nzVirtualMinBufferPx]="300"
  [nzVirtualForTrackBy]="trackByKey"
  [nzScroll]="{ y: '200px' }"
>
  <tbody>
    <tr *ngFor="let config of listOfData">
      <td>
        <strong>{{ config.key }}</strong>
      </td>
      <td class="td-value" [attr.title]="config.value">{{ config.value }}</td>
    </tr>
  </tbody>
</nz-table>
