<!--
  ~ Licensed to the Apache Software Foundation (ASF) under one
  ~ or more contributor license agreements.  See the NOTICE file
  ~ distributed with this work for additional information
  ~ regarding copyright ownership.  The ASF licenses this file
  ~ to you under the Apache License, Version 2.0 (the
  ~ "License"); you may not use this file except in compliance
  ~ with the License.  You may obtain a copy of the License at
  ~
  ~     http://www.apache.org/licenses/LICENSE-2.0
  ~
  ~ Unless required by applicable law or agreed to in writing, software
  ~ distributed under the License is distributed on an "AS IS" BASIS,
  ~ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  ~ See the License for the specific language governing permissions and
  ~ limitations under the License.
  -->

<nz-card nzType="inner" nzTitle="Execution Configuration">
  <nz-table
    *ngIf="config"
    class="no-border small"
    [nzTemplateMode]="true"
    [nzSize]="'small'"
    [nzFrontPagination]="false"
    [nzShowPagination]="false"
    [nzWidthConfig]="['50%', '50%']"
  >
    <tbody>
      <tr>
        <td><strong>Max. number of execution retries</strong></td>
        <td>{{ config['execution-config']['restart-strategy'] }}</td>
      </tr>
      <tr>
        <td><strong>Job parallelism</strong></td>
        <td>
          {{
            config['execution-config']['job-parallelism'] === -1
              ? 'auto'
              : config['execution-config']['job-parallelism']
          }}
        </td>
      </tr>
      <tr>
        <td><strong>Object reuse mode</strong></td>
        <td>{{ config['execution-config']['object-reuse-mode'] }}</td>
      </tr>
    </tbody>
  </nz-table>
</nz-card>
<nz-card nzType="inner" nzTitle="User Configuration">
  <nz-table
    #userTable
    *ngIf="config"
    class="no-border small"
    [nzData]="listOfUserConfig"
    [nzSize]="'small'"
    [nzFrontPagination]="false"
    [nzShowPagination]="false"
    [nzWidthConfig]="['50%', '50%']"
  >
    <tbody>
      <tr *ngFor="let config of userTable.data">
        <td>{{ config.key }}</td>
        <td>{{ config.value }}</td>
      </tr>
    </tbody>
  </nz-table>
</nz-card>
