<!--
  ~ Licensed to the Apache Software Foundation (ASF) under one
  ~ or more contributor license agreements.  See the NOTICE file
  ~ distributed with this work for additional information
  ~ regarding copyright ownership.  The ASF licenses this file
  ~ to you under the Apache License, Version 2.0 (the
  ~ "License"); you may not use this file except in compliance
  ~ with the License.  You may obtain a copy of the License at
  ~
  ~     http://www.apache.org/licenses/LICENSE-2.0
  ~
  ~ Unless required by applicable law or agreed to in writing, software
  ~ distributed under the License is distributed on an "AS IS" BASIS,
  ~ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  ~ See the License for the specific language governing permissions and
  ~ limitations under the License.
  -->

<div class="chart">
  <div class="title">
    <div class="text" [attr.title]="title">{{ title }}</div>
    <div class="operate">
      <nz-button-group [nzSize]="'small'">
        <button nz-button [nzType]="size === 'big' ? 'primary' : 'default'" (click)="resize('big')">
          Big
        </button>
        <button
          nz-button
          [nzType]="size === 'small' ? 'primary' : 'default'"
          (click)="resize('small')"
        >
          Small
        </button>
      </nz-button-group>
      <button nz-button class="close" [nzSize]="'small'" (click)="close()">
        <i nz-icon nzType="close" nzTheme="outline"></i>
      </button>
    </div>
  </div>
  <div class="content">
    <div [hidden]="displayMode === 'numeric'" #chart></div>
    <div class="numeric" [hidden]="displayMode === 'chart'">
      {{ latestValue | humanizeChartNumeric: title }}
    </div>
    <div class="type-switch">
      <nz-button-group [nzSize]="'small'">
        <button
          nz-button
          [nzType]="displayMode === 'chart' ? 'primary' : 'default'"
          (click)="setMode('chart')"
        >
          Chart
        </button>
        <button
          nz-button
          [nzType]="displayMode === 'numeric' ? 'primary' : 'default'"
          (click)="setMode('numeric')"
        >
          Numeric
        </button>
      </nz-button-group>
    </div>
  </div>
</div>
