/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.connector.upserttest.sink;

import org.apache.flink.annotation.VisibleForTesting;

import java.util.Arrays;
import java.util.Map;

import static org.apache.flink.util.Preconditions.checkNotNull;

/**
 * This class is a slim wrapper around {@code byte[]} that defensively copies the array during
 * construction and retrieval to prevent its contents from changing. It also implements a
 * content-dependent {@link #hashCode} method to allow usage as keys in a {@link Map}.
 */
class ImmutableByteArrayWrapper {

    @VisibleForTesting final byte[] bytes;

    ImmutableByteArrayWrapper(byte[] bytes) {
        checkNotNull(bytes);
        this.bytes = bytes.clone();
    }

    /**
     * Returns a reference-free copy of the underlying byte[].
     *
     * @return the copied byte[]
     */
    byte[] array() {
        return bytes.clone();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }
        ImmutableByteArrayWrapper that = (ImmutableByteArrayWrapper) o;
        return Arrays.equals(bytes, that.bytes);
    }

    @Override
    public int hashCode() {
        return Arrays.hashCode(bytes);
    }
}
