/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.jobmaster.event;

import java.util.HashMap;
import java.util.Map;

import static org.apache.flink.util.Preconditions.checkNotNull;

/** A class hosts all the type ids of events. */
public class JobEvents {

    private static final Map<Class<? extends JobEvent>, Integer> jobEventTypeIdMapping =
            new HashMap<>();

    static {
        jobEventTypeIdMapping.put(ExecutionVertexFinishedEvent.class, 1);
        jobEventTypeIdMapping.put(ExecutionVertexResetEvent.class, 2);
        jobEventTypeIdMapping.put(ExecutionJobVertexInitializedEvent.class, 3);
        jobEventTypeIdMapping.put(ExecutionJobVertexFinishedEvent.class, 4);
    }

    public static int getTypeID(Class<? extends JobEvent> clazz) {
        return checkNotNull(jobEventTypeIdMapping.get(clazz));
    }
}
