/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.module;

import org.apache.flink.annotation.Internal;
import org.apache.flink.configuration.ConfigOption;
import org.apache.flink.configuration.ConfigOptions;
import org.apache.flink.table.factories.Factory;

/** A collection of {@link ConfigOption} which are consistently used in multiple modules. */
@Internal
public class CommonModuleOptions {

    /**
     * {@link ConfigOption} which is used during module discovery to match it against {@link
     * Factory#factoryIdentifier()}.
     *
     * @deprecated This is only required for the legacy factory stack
     */
    @Deprecated
    public static final ConfigOption<String> MODULE_TYPE =
            ConfigOptions.key("type").stringType().noDefaultValue();
}
