/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.state.forst;

import org.apache.flink.core.asyncprocessing.InternalAsyncFuture;

import org.forstdb.RocksDB;
import org.forstdb.RocksDBException;

import java.io.IOException;
import java.util.Collection;

/**
 * The Put access request for ForStDB.
 *
 * @param <K> The type of key in put access request.
 * @param <N> The type of namespace in put access request.
 * @param <V> The type of value in put access request.
 */
public class ForStDBMultiRawMergePutRequest<K, N, V> extends ForStDBPutRequest<K, N, V> {

    final Collection<byte[]> rawValue;

    ForStDBMultiRawMergePutRequest(
            ContextKey<K, N> key,
            Collection<byte[]> value,
            ForStInnerTable<K, N, V> table,
            InternalAsyncFuture<Void> future) {
        super(key, null, true, table, future);
        this.rawValue = value;
    }

    @Override
    public void process(ForStDBWriteBatchWrapper writeBatchWrapper, RocksDB db)
            throws IOException, RocksDBException {
        byte[] key = buildSerializedKey();
        for (byte[] value : rawValue) {
            writeBatchWrapper.merge(table.getColumnFamilyHandle(), key, value);
        }
    }
}
