/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.client.cli.util;

import org.apache.flink.client.cli.CustomCommandLine;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.configuration.DeploymentOptions;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.Options;

/** Dummy implementation of the {@link CustomCommandLine} for testing purposes. */
public class DummyCustomCommandLine implements CustomCommandLine {

    @Override
    public boolean isActive(CommandLine commandLine) {
        return true;
    }

    @Override
    public String getId() {
        return DummyClusterClientFactory.ID;
    }

    @Override
    public void addRunOptions(Options baseOptions) {
        // nothing to add
    }

    @Override
    public void addGeneralOptions(Options baseOptions) {
        // nothing to add
    }

    @Override
    public Configuration toConfiguration(CommandLine commandLine) {
        final Configuration configuration = new Configuration();
        configuration.set(DeploymentOptions.TARGET, DummyClusterClientFactory.ID);
        return configuration;
    }
}
