/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.flink.runtime.scheduler;

import org.apache.flink.runtime.jobmanager.slots.TaskManagerGateway;
import org.apache.flink.runtime.jobmaster.TestingLogicalSlotBuilder;

/** Factory for {@link TestExecutionSlotAllocatorFactory}. */
public class TestExecutionSlotAllocatorFactory implements ExecutionSlotAllocatorFactory {

    private final TestExecutionSlotAllocator testExecutionSlotAllocator;

    private ExecutionSlotAllocationContext latestExecutionSlotAllocationContext;

    public TestExecutionSlotAllocatorFactory() {
        this.testExecutionSlotAllocator = new TestExecutionSlotAllocator();
    }

    public TestExecutionSlotAllocatorFactory(TaskManagerGateway gateway) {
        this.testExecutionSlotAllocator = new TestExecutionSlotAllocator(gateway);
    }

    public TestExecutionSlotAllocatorFactory(TestingLogicalSlotBuilder logicalSlotBuilder) {
        this.testExecutionSlotAllocator = new TestExecutionSlotAllocator(logicalSlotBuilder);
    }

    @Override
    public ExecutionSlotAllocator createInstance(final ExecutionSlotAllocationContext context) {
        this.latestExecutionSlotAllocationContext = context;
        return testExecutionSlotAllocator;
    }

    public TestExecutionSlotAllocator getTestExecutionSlotAllocator() {
        return testExecutionSlotAllocator;
    }

    public ExecutionSlotAllocationContext getLatestExecutionSlotAllocationContext() {
        return latestExecutionSlotAllocationContext;
    }
}
