/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
@import "theme";

:host {
  display: flex;
  flex-direction: column;
  height: 100%;

  ::ng-deep {
    .ant-divider-horizontal {
      margin: 0;
    }
  }

  .name {
    overflow: auto;
    max-height: 120px;
    padding: @padding-xs @padding-sm;
    font-size: @font-size-sm;
    white-space: pre-wrap;
    word-break: break-word;
  }

  .wrapper {
    flex: 1;

    dl {
      display: grid;
      grid-template-columns: auto 1fr;
      grid-gap: 0.75rem 2rem;
      margin-bottom: 0;
      padding: @padding-lg;
      font-size: @font-size-sm;

      dd {
        font-weight: 500;
      }

      dt {
        position: relative;
        font-weight: normal;

        &::before {
          content: ":";
          position: absolute;
          left: -1rem;
        }
      }
    }
  }
}
