/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

@import "theme";

:host {
  display: flex;
  flex: 1;
  flex-direction: column;

  .content {
    flex: 0 1 auto;
    box-sizing: border-box;
    margin-top: @margin-md;
    padding: 0;
    border: 1px solid @border-color-split;
    border-radius: 2px;
    background: @component-background;
    list-style: none;
    font-size: @font-size-base;
    line-height: 1.5;

    nz-skeleton {
      padding: @padding-md @padding-lg;
    }

    .router {
      display: flex;
      flex: 1;
      flex-flow: column nowrap;
    }
  }
}
