/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.state.forst;

import org.forstdb.ColumnFamilyHandle;
import org.forstdb.RocksDB;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.RegisterExtension;

/** Validate ForSt properties. */
class ForStPropertyTest {

    @RegisterExtension public ForStExtension forStExtension = new ForStExtension();

    @Test
    void testForStPropertiesValid() {
        RocksDB db = forStExtension.getDB();
        ColumnFamilyHandle handle = forStExtension.getDefaultColumnFamily();

        for (ForStProperty forStProperty : ForStProperty.values()) {
            try {
                forStProperty.getNumericalPropertyValue(db, handle);
            } catch (Exception e) {
                throw new AssertionError(
                        String.format(
                                "Invalid ForSt property %s", forStProperty.getForStProperty()),
                        e);
            }
        }
    }
}
