/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.flink.table.api.typeutils

import org.apache.flink.api.common.typeinfo.BasicTypeInfo
import org.apache.flink.api.common.typeutils.TypeInformationTestBase

/** Test for [[EitherTypeInfo]]. */
class EitherTypeInfoTest extends TypeInformationTestBase[EitherTypeInfo[_, _, _]] {

  override protected def getTestData: Array[EitherTypeInfo[_, _, _]] = Array(
    new EitherTypeInfo[Integer, Integer, Either[Integer, Integer]](
      classOf[Either[Integer, Integer]],
      BasicTypeInfo.INT_TYPE_INFO,
      BasicTypeInfo.INT_TYPE_INFO
    ),
    new EitherTypeInfo[Integer, String, Either[Integer, String]](
      classOf[Either[Integer, String]],
      BasicTypeInfo.INT_TYPE_INFO,
      BasicTypeInfo.STRING_TYPE_INFO
    )
  )

}
