/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.planner.operations.converters;

import org.apache.flink.sql.parser.ddl.SqlAlterModelReset;
import org.apache.flink.table.api.ValidationException;
import org.apache.flink.table.catalog.ModelChange;
import org.apache.flink.table.catalog.ResolvedCatalogModel;
import org.apache.flink.table.catalog.UnresolvedIdentifier;
import org.apache.flink.table.operations.Operation;
import org.apache.flink.table.operations.ddl.AlterModelChangeOperation;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

/** A converter for {@link org.apache.flink.sql.parser.ddl.SqlAlterModelReset}. */
public class SqlAlterModelResetConverter
        extends AbstractSqlAlterModelConverter<SqlAlterModelReset> {

    @Override
    public Operation convertSqlNode(SqlAlterModelReset sqlAlterModelReset, ConvertContext context) {
        ResolvedCatalogModel existingModel =
                getExistingModel(
                        context,
                        sqlAlterModelReset.fullModelName(),
                        sqlAlterModelReset.ifModelExists());

        Set<String> resetKeys = sqlAlterModelReset.getResetKeys();
        if (resetKeys.isEmpty()) {
            throw new ValidationException("ALTER MODEL RESET does not support empty key.");
        }
        List<ModelChange> modelChanges =
                resetKeys.stream().map(ModelChange::reset).collect(Collectors.toList());

        if (existingModel == null) {
            return new AlterModelChangeOperation(
                    context.getCatalogManager()
                            .qualifyIdentifier(
                                    UnresolvedIdentifier.of(sqlAlterModelReset.fullModelName())),
                    modelChanges,
                    null,
                    sqlAlterModelReset.ifModelExists());
        }

        Map<String, String> newOptions = new HashMap<>(existingModel.getOptions());
        resetKeys.forEach(newOptions::remove);

        return new AlterModelChangeOperation(
                context.getCatalogManager()
                        .qualifyIdentifier(
                                UnresolvedIdentifier.of(sqlAlterModelReset.fullModelName())),
                modelChanges,
                existingModel.copy(newOptions),
                sqlAlterModelReset.ifModelExists());
    }
}
