/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.taskexecutor;

import org.apache.flink.runtime.executiongraph.ExecutionAttemptID;

import java.io.Serializable;
import java.util.Set;

/** A report about the currently deployed executions of a TaskExecutor. */
public class ExecutionDeploymentReport implements Serializable {

    private static final long serialVersionUID = -2731996813330298044L;

    private final Set<ExecutionAttemptID> executions;

    public ExecutionDeploymentReport(Set<ExecutionAttemptID> executions) {
        this.executions = executions;
    }

    public Set<ExecutionAttemptID> getExecutions() {
        return executions;
    }

    @Override
    public String toString() {
        return "ExecutionDeploymentReport{" + "executions=" + executions + '}';
    }
}
