/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.api.common.io;

import org.apache.flink.annotation.Public;
import org.apache.flink.core.io.IOReadableWritable;
import org.apache.flink.core.memory.DataInputView;

import java.io.IOException;

/**
 * Reads elements by deserializing them with their regular serialization/deserialization
 * functionality.
 *
 * @see SerializedOutputFormat
 */
@Public
public class SerializedInputFormat<T extends IOReadableWritable> extends BinaryInputFormat<T> {

    private static final long serialVersionUID = 1L;

    @Override
    protected T deserialize(T reuse, DataInputView dataInput) throws IOException {
        reuse.read(dataInput);
        return reuse;
    }
}
