/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

@import "theme";

:host {
  display: flex;
  flex: 1;
  flex-flow: column nowrap;

  ::ng-deep {
    .ant-tabs-nav-list {
      padding: 4px 16px;
    }
  }

  nz-tabset {
    position: relative;
    flex: 1;
  }

  nz-code-editor {
    position: relative;
    height: calc(~"100vh - 310px");

    &.subtask {
      position: relative;
      top: 0;
      height: 300px;
    }
  }

  nz-table {
    margin-top: -@margin-md;
  }

  nz-tag {
    margin-bottom: 5px;
  }

  .expand-td {
    display: block;
    padding: 0 !important;
  }

  .exception-select {
    width: 300px;
  }
}
