/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.rest.handler.legacy.metrics;

/**
 * The MetricFetcher can be used to fetch metrics from the JobManager and all registered
 * TaskManagers.
 *
 * <p>Metrics will only be fetched when {@link MetricFetcher#update()} is called, provided that a
 * sufficient time since the last call has passed.
 */
public interface MetricFetcher {

    /**
     * Get {@link MetricStore} which contains all currently fetched metrics.
     *
     * @return {@link MetricStore} with all fetched metrics
     */
    MetricStore getMetricStore();

    /** Trigger fetching of metrics. */
    void update();

    /**
     * @return timestamp of the last update.
     */
    long getLastUpdateTime();
}
