/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
@import "theme";

:host {
  display: flex;
  flex: 1;
  flex-direction: column;
  overflow: auto;

  ::ng-deep {
    .ant-table-cell {
      font-size: @font-size-sm;
    }
  }

  .header-icon {
    position: relative;
    top: 1px;
  }

  nz-progress {
    display: block;
    width: 90%;

    ::ng-deep {
      .ant-progress-inner {
        border-radius: 0;
      }
    }
  }

  .flink-memory-model {
    margin-bottom: @margin-md;

    .table-header {
      font-weight: 500;
    }

    ::ng-deep {
      .ant-table-small {
        tr:hover {
          td,
          th {
            background: transparent;
          }
        }
      }
    }
  }

  .top-margin {
    margin-top: @margin-md;
  }
}
