/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.factories;

import org.apache.flink.annotation.Internal;
import org.apache.flink.table.legacy.factories.TableFactory;

import java.util.List;
import java.util.Map;

/**
 * Exception for finding more than one {@link TableFactory} for the given properties.
 *
 * <p>This replaces the deprecated {@link
 * org.apache.flink.table.api.AmbiguousTableFactoryException}.
 */
@Internal
public class AmbiguousTableFactoryException
        extends org.apache.flink.table.api.AmbiguousTableFactoryException {

    public AmbiguousTableFactoryException(
            List<? extends TableFactory> matchingFactories,
            Class<? extends TableFactory> factoryClass,
            List<TableFactory> factories,
            Map<String, String> properties,
            Throwable cause) {
        super(matchingFactories, factoryClass, factories, properties, cause);
    }

    public AmbiguousTableFactoryException(
            List<? extends TableFactory> matchingFactories,
            Class<? extends TableFactory> factoryClass,
            List<TableFactory> factories,
            Map<String, String> properties) {
        super(matchingFactories, factoryClass, factories, properties);
    }
}
