/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.runtime.operators.process;

import org.apache.flink.table.types.DataType;
import org.apache.flink.table.types.inference.TypeInferenceUtil.StateInfo;

import java.io.Serializable;

/**
 * Serializable representation of {@link StateInfo} that will translate into state descriptors and
 * state access.
 */
public class RuntimeStateInfo implements Serializable {

    private static final long serialVersionUID = 1L;

    private final String stateName;
    private final DataType dataType;
    private final long timeToLive;

    public RuntimeStateInfo(String stateName, DataType dataType, long timeToLive) {
        this.stateName = stateName;
        this.dataType = dataType;
        this.timeToLive = timeToLive;
    }

    public String getStateName() {
        return stateName;
    }

    public DataType getDataType() {
        return dataType;
    }

    public long getTimeToLive() {
        return timeToLive;
    }
}
