/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.core.fs;

import org.apache.flink.annotation.Internal;
import org.apache.flink.util.Preconditions;

import java.io.IOException;
import java.nio.ByteBuffer;

/** A {@link ClosingFSDataInputStream} which is {@link ByteBufferReadable}. */
@Internal
public class ByteBufferReadableClosingFSDataInputStream extends ClosingFSDataInputStream
        implements ByteBufferReadable {

    private final ByteBufferReadable inputStream;

    protected ByteBufferReadableClosingFSDataInputStream(
            FSDataInputStream delegate, SafetyNetCloseableRegistry registry, String debugInfo)
            throws IOException {
        super(delegate, registry, debugInfo);
        Preconditions.checkArgument(delegate instanceof ByteBufferReadable);
        this.inputStream = (ByteBufferReadable) delegate;
    }

    @Override
    public int read(ByteBuffer byteBuffer) throws IOException {
        return inputStream.read(byteBuffer);
    }

    @Override
    public int read(long position, ByteBuffer byteBuffer) throws IOException {
        return inputStream.read(position, byteBuffer);
    }
}
