/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.state.heap;

import org.apache.flink.runtime.state.InternalKeyContext;
import org.apache.flink.runtime.state.InternalKeyContextImpl;
import org.apache.flink.runtime.state.KeyGroupRange;
import org.apache.flink.runtime.state.KeyGroupRangeAssignment;

/** Mock {@link InternalKeyContext}. */
public class MockInternalKeyContext<K> extends InternalKeyContextImpl<K> {

    MockInternalKeyContext() {
        super(new KeyGroupRange(0, 0), 1);
    }

    MockInternalKeyContext(int startKeyGroup, int endKeyGroup, int numberOfKeyGroups) {
        super(new KeyGroupRange(startKeyGroup, endKeyGroup), numberOfKeyGroups);
    }

    public void setCurrentKeyAndKeyGroup(K key) {
        super.setCurrentKey(key);
        super.setCurrentKeyGroupIndex(
                KeyGroupRangeAssignment.assignToKeyGroup(key, getNumberOfKeyGroups()));
    }
}
