/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.state.v2;

import org.apache.flink.api.common.state.v2.StateIterator;
import org.apache.flink.runtime.asyncprocessing.AbstractStateIterator;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.NoSuchElementException;

public class SyncIteratorWrapper<T> implements Iterator<T> {
    private final ArrayList<T> cacheEntries;
    private int cacheIndex;

    public SyncIteratorWrapper(StateIterator<T> stateIterator) {
        this.cacheEntries = new ArrayList<>();
        this.cacheIndex = 0;
        if (stateIterator instanceof AbstractStateIterator) {
            ((AbstractStateIterator<T>) stateIterator).onNextSync(this.cacheEntries::add);
        }
    }

    @Override
    public boolean hasNext() {
        return (cacheIndex < cacheEntries.size());
    }

    @Override
    public T next() {
        if (cacheIndex == cacheEntries.size()) {
            throw new NoSuchElementException("Iterator has no more elements!");
        }
        T currentEntry = cacheEntries.get(cacheIndex);
        cacheIndex++;
        return currentEntry;
    }
}
