/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
@import "theme";

:host {
  display: block;
  width: 100%;
  height: 100%;

  .text-secondary {
    color: @text-color-secondary;
    font-size: @font-size-sm;
  }

  nz-tabset {
    height: 100%;
  }

  ::ng-deep {
    .ant-table-row-indent {
      padding-left: 0 !important;

      & + .ant-table-row-expand-icon {
        margin-top: 2px;
      }
    }

    .ant-table-cell {
      font-size: @font-size-sm;
    }

    .ant-tabs-content {
      height: 100%;

      .ant-tabs-tabpane {
        height: 100%;
      }
    }

    .ant-tabs-nav {
      margin-bottom: 0;
    }

    .ant-tabs-small > .ant-tabs-nav .ant-tabs-tab {
      margin: 0 0 0 @margin-md;
      font-size: @font-size-sm;
    }

    ::-webkit-scrollbar {
      display: none;
    }

    nz-table,
    nz-spin,
    cdk-virtual-scroll-viewport,
    nz-table-inner-scroll,
    .ant-spin-container,
    .ant-table {
      height: 100%;
    }
  }
}
