/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.api.internal;

import org.apache.flink.annotation.Internal;
import org.apache.flink.api.dag.Transformation;
import org.apache.flink.table.operations.CollectModifyOperation;
import org.apache.flink.table.operations.QueryOperation;

import java.util.List;

/** PlanCache for DQL. */
@Internal
public class DQLCachedPlan implements CachedPlan {
    private final QueryOperation queryOperation;
    private final CollectModifyOperation sinkOperation;
    private final List<Transformation<?>> transformations;

    public DQLCachedPlan(
            QueryOperation queryOperation,
            CollectModifyOperation sinkOperation,
            List<Transformation<?>> transformations) {
        this.queryOperation = queryOperation;
        this.sinkOperation = sinkOperation;
        this.transformations = transformations;
    }

    @Override
    public QueryOperation getOperation() {
        return queryOperation;
    }

    public CollectModifyOperation getSinkOperation() {
        return sinkOperation;
    }

    public List<Transformation<?>> getTransformations() {
        return transformations;
    }
}
