/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.runtime.util;

import org.apache.flink.table.runtime.operators.join.FlinkJoinType;

import static org.apache.flink.table.runtime.operators.join.FlinkJoinType.FULL;
import static org.apache.flink.table.runtime.operators.join.FlinkJoinType.INNER;
import static org.apache.flink.table.runtime.operators.join.FlinkJoinType.RIGHT;

/** Utility for join. */
public class JoinUtil {

    public static FlinkJoinType getJoinType(boolean leftOuter, boolean rightOuter) {
        if (leftOuter && rightOuter) {
            return FULL;
        } else if (leftOuter) {
            return FlinkJoinType.LEFT;
        } else if (rightOuter) {
            return RIGHT;
        } else {
            return INNER;
        }
    }

    private JoinUtil() {}
}
