/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.apache.flink.table.gateway.service.utils;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/** Handler to log the exception and exits. */
public class IgnoreExceptionHandler implements Thread.UncaughtExceptionHandler {

    public static final IgnoreExceptionHandler INSTANCE = new IgnoreExceptionHandler();

    private static final Logger LOG = LoggerFactory.getLogger(IgnoreExceptionHandler.class);

    @Override
    public void uncaughtException(Thread t, Throwable e) {
        // ignore error
        LOG.error("Thread '{}' produced an uncaught exception. Ignore...", t.getName(), e);
    }
}
