/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.jdbc.utils;

import org.apache.flink.table.client.gateway.StatementResult;
import org.apache.flink.table.data.RowData;

/** Closeable result iterator for statement result. */
public class StatementResultIterator implements CloseableResultIterator<RowData> {
    private final StatementResult result;

    public StatementResultIterator(StatementResult result) {
        this.result = result;
    }

    @Override
    public void close() throws Exception {
        result.close();
    }

    @Override
    public boolean hasNext() {
        return result.hasNext();
    }

    @Override
    public RowData next() {
        return result.next();
    }
}
