/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.planner.expressions.converter;

import org.apache.flink.annotation.Internal;
import org.apache.flink.table.expressions.CallExpression;
import org.apache.flink.table.planner.expressions.converter.converters.CustomizedConverters;

import org.apache.calcite.rex.RexNode;

import java.util.Optional;

/**
 * Customized {@link CallExpressionConvertRule}, Functions conversion here all require special
 * logic, and there may be some special rules, such as needing get the literal values of inputs,
 * such as converting to combinations of functions, to convert to RexNode of calcite.
 */
@Internal
public class CustomizedConvertRule implements CallExpressionConvertRule {

    private final CustomizedConverters customizedConverters = new CustomizedConverters();

    @Override
    public Optional<RexNode> convert(CallExpression call, ConvertContext context) {
        return customizedConverters
                .getConverter(call.getFunctionDefinition())
                .map(converter -> converter.convert(call, context));
    }
}
